/**CFile****************************************************************

  FileName    [extra.h]

  PackageName [extra]

  Synopsis    [Experimental version of some DD-based procedures.]

  Description [This library contains a number of operators and 
  traversal routines developed to extend the functionality of 
  CUDD v.2.3.x, by Fabio Somenzi (http://vlsi.colorado.edu/~fabio/)
  To compile your code with the library, #include "extra.h" 
  in your source files and link your project to CUDD and this 
  library. Use the library at your own risk and with caution. 
  Notice that debugging of some operators still continues.]

  Author      [Alan Mishchenko]
  
  Affiliation [UC Berkeley]

  Date        [Ver. 1.0. Started - February 1, 2003.]

  Revision    [$Id: extra.h,v 1.9 2003/05/27 23:14:36 alanmi Exp $]

***********************************************************************/

#ifndef __EXTRA_H__
#define __EXTRA_H__

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

#include <string.h>
#include <time.h>
#include "util.h"
#include "st.h"
#include "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

// an assuption about the maximum number of BDD variables
#define MAXINPUTS   1000       

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/* constants of the manager */
#define		b0     Cudd_Not((dd)->one)
#define		b1              (dd)->one
#define		z0              (dd)->zero
#define		z1              (dd)->one
#define		a0              (dd)->zero
#define		a1              (dd)->one


// hash key macros
#define hashKey1(a,TSIZE) \
((unsigned)(a) % TSIZE)

#define hashKey2(a,b,TSIZE) \
(((unsigned)(a) + (unsigned)(b) * DD_P1) % TSIZE)

#define hashKey3(a,b,c,TSIZE) \
(((((unsigned)(a) + (unsigned)(b)) * DD_P1 + (unsigned)(c)) * DD_P2 ) % TSIZE)

#define hashKey4(a,b,c,d,TSIZE) \
((((((unsigned)(a) + (unsigned)(b)) * DD_P1 + (unsigned)(c)) * DD_P2 + \
   (unsigned)(d)) * DD_P3) % TSIZE)

#define hashKey5(a,b,c,d,e,TSIZE) \
(((((((unsigned)(a) + (unsigned)(b)) * DD_P1 + (unsigned)(c)) * DD_P2 + \
   (unsigned)(d)) * DD_P3 + (unsigned)(e)) * DD_P1) % TSIZE)


// complementation and testing for pointers for hash entries
#define Hash_IsComplement(p)  ((int)((long) (p) & 02))
#define Hash_Regular(p)       ((DdNode*)((unsigned)(p) & ~02))
#define Hash_Not(p)           ((DdNode*)((long)(p) ^ 02))
#define Hash_NotCond(p,c)     ((DdNode*)((long)(p) ^ (02*c)))

// print the BDD of the function in the SOP form
#define PRB(dd,f)       printf("%s = ", #f); Extra_bddPrint(dd,f); printf("\n")
// print the ZDD of the function in the SOP form
#define PRZ(dd,f)       printf("%s = \n", #f); Cudd_zddPrintCover(dd,f); 
// print the number of nodes in the BDD
#define PRN(f)          printf("Nodes in %s = %d\n", #f, Cudd_DagSize(f))
// print the support of the BDD
#define PRS(dd,f)       printf("Support of %s = ", #f ); Extra_PrintSupport((dd),(f)); printf("\n")
// print the Karnaugh map of the completely-specified function
#define PRK(dd,f,n)     printf("K-map for function <%s> is:\n", #f); Extra_PrintKMap(stdout,dd,(f),Cudd_Not(f),(n),NULL,0)
// print the Karnaugh map of the incompletely-specified function
#define PRK2(dd,f,g,n)  printf("K-map for function <%s> is:\n", #f); Extra_PrintKMap(stdout,dd,(f),(g),(n),NULL,0)
// print the time spent by the process "a"
#define PRT(a,t)        printf("%s = ", (a)); printf( "%6.2f sec\n", (float)(t)/(float)(CLOCKS_PER_SEC) )

#define CALLOC(type, num)  ((type *) calloc((long)(num), (long)sizeof(type)))

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/


/*===========================================================================*/
/*    ADD-based procedures                                                   */
/*===========================================================================*/

/*=== extraAddMisc.c =============================================================*/

/* remaps the function to depend on the topmost variables on the manager. */
extern DdNode *     Extra_addRemapUp ARGS((DdManager * dd, DdNode * aF));
/* counts the number of different constant nodes of the ADD */
extern int          Extra_addCountConst ARGS((DdManager * dd, DdNode * aFunc));
/* counts the the number of different constant nodes of the array of ADDs */
extern int          Extra_addCountConstArray ARGS((DdManager * dd, DdNode ** paFuncs, int nFuncs));
/* finds the minimum value terminal node in the array of ADDs */
extern DdNode *     Extra_addFindMinArray ARGS((DdManager * dd, DdNode ** paFuncs, int nFuncs));
/* finds the maximum value terminal node in the array of ADDs */
extern DdNode *     Extra_addFindMaxArray ARGS((DdManager * dd, DdNode ** paFuncs, int nFuncs));
/* absolute value of an ADD */
extern DdNode *     Extra_addAbs ARGS((DdManager * dd, DdNode * f));
/* the encoded set of absolute values of the constant nodes of an ADD */
extern DdNode *     Extra_bddAddConstants ARGS((DdManager * dd, DdNode * aFunc));
/* determines whether this is an ADD or a BDD */
extern int          Extra_addDetectAdd ARGS((DdManager * dd, DdNode * Func));
/* restructure the ADD by replacing negative terminals with their abs value */
extern DdNode *     Cudd_addAbs ARGS((DdManager * dd, DdNode * f));
/* Coudert's restrict applicable to true ADDs (takes the don't-care set, not the care set). */
extern DdNode *     Extra_addRestrictAdd ARGS((DdManager * dd, DdNode * aF, DdNode * aD));
extern DdNode *      extraAddRestrictAdd ARGS((DdManager * dd, DdNode * aF, DdNode * aD));
/* intersection, union, and sharp of paths for each terminal */
extern DdNode *     Extra_addForeachTerminalAnd ARGS((DdManager * dd, DdNode ** f, DdNode ** g));
extern DdNode *     Extra_addForeachTerminalOr ARGS((DdManager * dd, DdNode ** f, DdNode ** g));
extern DdNode *     Extra_addForeachTerminalSharp ARGS((DdManager * dd, DdNode ** f, DdNode ** g));
/* swapping a given terminal node with zero terminal node */
extern DdNode *     Extra_addSwapTerminals ARGS((DdManager * dd, DdNode * aFunc, DdNode * aTerm));
extern DdNode *      extraAddSwapTerminals ARGS((DdManager * dd, DdNode * aFunc, DdNode * aTerm));

/*=== extraAddSpectra.c =============================================================*/

/* Walsh spectrum computation */
extern DdNode *     Extra_bddWalsh ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
extern DdNode *      extraBddWalsh ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
/* Reed-Muller spectrum computation */
extern DdNode *     Extra_bddReedMuller ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
extern DdNode *      extraBddReedMuller ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
/* Haar spectrum computation */
extern DdNode *     Extra_bddHaar ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
extern DdNode *      extraBddHaar ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
/* inverse Haar spectrum computation */
extern DdNode *     Extra_bddHaarInverse ARGS((DdManager * dd, DdNode * aFunc, DdNode * bVars));
extern DdNode *      extraBddHaarInverse ARGS((DdManager * dd, DdNode * aFunc, DdNode * aSteps, DdNode * bVars, DdNode * bVarsAll, int nVarsAll, int * InverseMap));
/* remapping from the natural order into the sequential order */
extern DdNode *     Extra_addRemapNatural2Sequential ARGS((DdManager * dd, DdNode * aSource, DdNode * bVars));
/* addition function to update the negative-variable-assignment cube value in the ADD */
extern DdNode *      extraAddUpdateZeroCubeValue ARGS((DdManager * dd, DdNode * aFunc, DdNode * bVars, DdNode * aNode));
/* the generalized ITE for ADDs */
extern DdNode *      extraAddIteRecurGeneral ARGS((DdManager * dd, DdNode * bX, DdNode * aF, DdNode * aG));



/*===========================================================================*/
/*    BDD-based procedures                                                   */
/*===========================================================================*/

/*=== extraBddKmap.c ================================================================*/

/* displays the Karnaugh Map of a function */
extern void        Extra_PrintKMap ARGS((FILE * pFile, DdManager * dd, DdNode * OnSet, DdNode * OffSet, int nVars, DdNode ** XVars, int fSuppType, char ** pVarNames));
/* displays the Karnaugh Map of a relation */
extern void        Extra_PrintKMapRelation ARGS((FILE * pFile, DdManager * dd, DdNode * OnSet, DdNode * OffSet, int nXVars, int nYVars, DdNode ** XVars, DdNode ** YVars));

/*=== extraBddMisc.c ===============================================================*/

/* remaps the function to depend on the topmost variables on the manager. */
extern DdNode *    Extra_bddRemapUp ARGS((DdManager * dd, DdNode * bF));
/* finds one cube belonging to the on-set of the function */
extern DdNode *    Extra_bddFindOneCube ARGS((DdManager * dd, DdNode * bF));
/* finds one minterm belonging to the on-set of the function */
extern DdNode *    Extra_bddFindOneMinterm ARGS((DdManager * dd, DdNode * bF, int nVars));
/* finds one cube belonging to the on-set, with the smallest bits-to-integer value */
extern DdNode *    Extra_bddGetOneCube ARGS((DdManager * dd, DdNode * bFunc));
/* finds one minterm belonging to the on-set, with the smallest bits-to-integer value */
extern DdNode *    Extra_bddGetOneMinterm ARGS((DdManager * dd, DdNode * bFunc, DdNode * bSupp));

/* compute the random function with the given density */
extern DdNode *    Extra_bddRandomFunc ARGS((DdManager * dd, int n, double d));
/* prints the bdd in the form of disjoint sum of products */
extern void        Extra_bddPrint ARGS((DdManager * dd, DdNode * F));
/* build the set of all tuples of K variables out of N */
extern DdNode *    Extra_bddTuples ARGS((DdManager * dd, int K, DdNode *bVarsN));
extern DdNode *     extraBddTuples ARGS((DdManager * dd, DdNode *bVarsK, DdNode *bVarsN));
/* changes the polarity of vars listed in the cube */
extern DdNode *    Extra_bddChangePolarity ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));
extern DdNode *     extraBddChangePolarity ARGS((DdManager * dd, DdNode * bFunc, DdNode * bVars));

/* converts the bit-string (nbits <= 32) into the BDD cube */
extern DdNode *    Extra_bddBitsToCube ARGS((DdManager * dd, int Code, int CodeWidth, DdNode ** pbVars));
/* computes the positive polarty cube composed of the first vars in the array */
extern DdNode *    Extra_bddComputeCube ARGS((DdManager * dd, DdNode ** bXVars, int nVars));
/* visualize the BDD/ADD/ZDD */
extern void        Extra_bddShow ARGS((DdManager * dd, DdNode * bFunc));
extern void        Extra_addShowFromBdd ARGS((DdManager * dd, DdNode * bFunc));
/* performs the boolean difference w.r.t to a cube (AKA unique quontifier) */
extern DdNode *    Extra_bddBooleanDiffCube ARGS((DdManager * dd, DdNode * bFunc, DdNode * bCube));
extern DdNode *     extraBddBooleanDiffCube ARGS((DdManager * dd, DdNode * bFunc, DdNode * bCube));


/* find the profile of a DD (the number of nodes on each level) */
extern int *       Extra_ProfileNode ARGS((DdManager * dd, DdNode * F, int * Profile));
/* find the profile of a shared set of DDs (the number of nodes on each level) */
extern int *       Extra_ProfileNodeSharing ARGS((DdManager * dd, DdNode ** pFuncs, int nFuncs, int * Profile));

/* find the profile of a DD (the number of edges of each length) */
extern int *       Extra_ProfileEdge ARGS((DdManager * dd, DdNode * F, int * Profile));
/* find the profile of a shared set of DDs (the number of edges of each length) */
extern int *       Extra_ProfileEdgeSharing ARGS((DdManager * dd, DdNode ** pFuncs, int nFuncs, int * Profile));
/* permutes variables the array of functions */
extern void        Extra_bddPermuteArray ARGS((DdManager * dd, DdNode ** bNodesIn, DdNode ** bNodesOut, int nNodes, int *permut));
/* permutes variables the BDD and returns an ADD */
extern DdNode *    Extra_bddPermuteToAdd ARGS((DdManager * dd, DdNode * bFunc, int * Permute));

/* finds the smallest integer larger of equal than the logarithm. */
extern int         Extra_Base2Log ARGS((unsigned Num));
extern int         Extra_Base2LogDouble ARGS((double Num));
/* returns the power of two as a double */
extern double      Extra_Power2 ARGS((unsigned Num));

/* checks if the DD has at least one complement edge */
extern int         Extra_WithComplementedEdges ARGS((DdNode * bFunc));

/*=== extraBddPermute.c =================================================================*/

typedef struct _permman  permman; // the permutation manager data structure

/* starting and stopping the permutation manager */
extern permman *   Extra_PermutationManagerInit ARGS(());
extern void        Extra_PermutationManagerQuit ARGS((permman * p));
/* set/reset the key increase limit */
extern void        Extra_PermutationManagerKeyIncreaseLimitSet ARGS((permman * p, int Limit));
extern void        Extra_PermutationManagerKeyIncreaseLimitReset ARGS((permman * p));

/* permuting the BDD in a way that is more efficient than Cudd_bddPermute() */
extern DdNode *    Extra_Permute ARGS((permman * pMan, DdManager * dd, DdNode * Func, int * pPermute));
extern DdNode *    Extra_RemapUp ARGS((permman * pMan, DdManager * dd, DdNode * Func));

/*=== extraBddSupp.c =================================================================*/

/* returns the size of the support */
extern int         Extra_bddSuppSize ARGS((DdManager * dd, DdNode * bSupp));
/* returns 1 if the support contains the given BDD variable */
extern int         Extra_bddSuppContainVar ARGS((DdManager * dd, DdNode * bS, DdNode * bVar));
/* returns 1 if two supports represented as BDD cubes are overlapping */
extern int         Extra_bddSuppOverlapping ARGS((DdManager * dd, DdNode * S1, DdNode * S2));
/* returns the number of different vars in two supports */
extern int         Extra_bddSuppDifferentVars ARGS((DdManager * dd, DdNode * S1, DdNode * S2, int DiffMax));
/* checks the support containment */
extern int         Extra_bddSuppCheckContainment ARGS((DdManager * dd, DdNode * bL, DdNode * bH, DdNode ** bLarge, DdNode ** bSmall));

/* get support of the DD as an array of integers */
extern int *       Extra_SupportArray ARGS((DdManager * dd, DdNode * F, int * support));
/* get support of the array of DDs as an array of integers */
extern int *       Extra_VectorSupportArray ARGS((DdManager * dd, DdNode ** F, int n, int * support));
/* get support size and node count at the same time */
extern int         Extra_DagSizeSuppSize ARGS((DdNode * node, int * pnSuppSize));
/* get support of support and stores it in cache */
extern DdNode *    Extra_SupportCache ARGS((DdManager * dd, DdNode * f));
/* get support of the DD as a ZDD */
extern DdNode *    Extra_zddSupport ARGS((DdManager * dd, DdNode * f));
/* get support as a negative polarity BDD cube */
extern DdNode *    Extra_bddSupportNegativeCube ARGS((DdManager * dd, DdNode * f));


/*===========================================================================*/
/*    Mixed-DD procedures                                                    */
/*===========================================================================*/

/*=== extraDdShift.c =================================================================*/

/* shifts the BDD up/down by one variable */
extern DdNode *    Extra_bddShift ARGS((DdManager * dd, DdNode * bF, int fShiftUp));
extern DdNode *     extraBddShift ARGS((DdManager * dd, DdNode * bF, DdNode * bFlag));
/* stretches the BDD the given number of times  */
extern DdNode *    Extra_bddStretch ARGS((DdManager * dd, DdNode * bF, int nTimes));
extern DdNode *     extraBddStretch ARGS((DdManager * dd, DdNode * bF, DdNode * bTimes));
/* shifts the ZDD up/down by one variable */
extern DdNode *    Extra_zddShift ARGS((DdManager * dd, DdNode * zF, int fShiftUp));
extern DdNode *     extraZddShift ARGS((DdManager * dd, DdNode * zF, DdNode * zFlag));
/* swaps two variables in the BDD */
extern DdNode *    Extra_bddSwapVars ARGS((DdManager * dd, DdNode * bF, int iVar1, int iVar2));
extern DdNode *     extraBddSwapVars ARGS((DdManager * dd, DdNode * bF, DdNode * bVars));

/*=== extraDdTimed.c =================================================================*/

/* sets the timeout and spaceout for Extra_bddAnd(), Extra_bddOr(), and Extra_bddBooleanDiff() */
extern void        Extra_OperationTimeoutSet( int timeout );
extern void        Extra_OperationSpaceoutSet( int MaxNodeIncrease );
extern void        Extra_OperationTimeoutReset();
extern void        Extra_OperationSpaceoutReset();

/* the same as normal operators only with timeout and spaceout */
extern DdNode *    Extra_bddAnd( DdManager * dd, DdNode * f, DdNode * g );
extern DdNode *    Extra_bddOr( DdManager * dd, DdNode * f, DdNode * g );
extern DdNode *     extraBddAndRecur( DdManager * manager, DdNode * f, DdNode * g );
extern DdNode *    Extra_bddBooleanDiff( DdManager * manager, DdNode * f, int x );
extern DdNode *     extraBddBooleanDiffRecur( DdManager * manager, DdNode * f, DdNode * var );
extern DdNode *    Extra_bddVectorCompose( DdManager * dd, DdNode * f, DdNode ** vector );
extern DdNode *    Extra_zddIsopCoverAltTimed( DdManager * dd, DdNode * bFuncOn, DdNode * bFuncOnDc );
extern DdNode *    Extra_zddIsopCoverAltLimited( DdManager * dd, DdNode * bFuncOn, DdNode * bFuncOnDc, int nBTracks );

/*=== extraDdTransfer.c =================================================================*/

/* convert a {A,B}DD from a manager to another with variable remapping */
extern DdNode *    Extra_TransferPermute ARGS((DdManager * ddSource, DdManager * ddDestination, DdNode * f, int * Permute));

/*=== extraDdUtils.c =============================================================*/

extern DdNode *    Extra_zddCoveredByArea ARGS((DdManager *dd, DdNode *zC, DdNode *bA));
extern DdNode *     extraZddCoveredByArea ARGS((DdManager *dd, DdNode *zC, DdNode *bA));
extern DdNode *    Extra_zddNotCoveredByCover ARGS((DdManager *dd, DdNode *zC, DdNode *zD));
extern DdNode *     extraZddNotCoveredByCover ARGS((DdManager *dd, DdNode *zC, DdNode *zD));
extern DdNode *    Extra_zddOverlappingWithArea ARGS((DdManager *dd, DdNode *zC, DdNode *bA));
extern DdNode *     extraZddOverlappingWithArea ARGS((DdManager *dd, DdNode *zC, DdNode *bA));
extern DdNode *    Extra_zddConvertToBdd ARGS((DdManager *dd, DdNode *zC));
extern DdNode *     extraZddConvertToBdd ARGS((DdManager *dd, DdNode *zC));
extern DdNode *    Extra_zddConvertToBddUnate ARGS((DdManager *dd, DdNode *zC));
extern DdNode *     extraZddConvertToBddUnate ARGS((DdManager *dd, DdNode *zC));
extern DdNode *    Extra_zddConvertEsopToBdd ARGS((DdManager *dd, DdNode *zC));
extern DdNode *     extraZddConvertEsopToBdd ARGS((DdManager *dd, DdNode *zC));
extern DdNode *    Extra_zddConvertToBddAndAdd ARGS((DdManager *dd, DdNode *zC, DdNode *bA));
extern DdNode *     extraZddConvertToBddAndAdd ARGS((DdManager *dd, DdNode *zC, DdNode *bA));
extern DdNode *    Extra_zddSingleCoveredArea ARGS((DdManager *dd, DdNode *zC));
extern DdNode *     extraZddSingleCoveredArea ARGS((DdManager *dd, DdNode *zC));
extern DdNode *    Extra_zddConvertBddCubeIntoZddCube ARGS((DdManager *dd, DdNode *bCube));
extern int         Extra_zddVerifyCover ARGS((DdManager * dd, DdNode * zC, DdNode * bFuncOn, DdNode * bFuncOnDc));



/*===========================================================================*/
/*    ZDD-based procedures                                                   */
/*===========================================================================*/

/*=== extraZddCovers.c ==============================================================*/

/* the result of this operation is primes contained in the product of cubes */
extern DdNode *    Extra_zddPrimeProduct ARGS((DdManager *dd, DdNode *f, DdNode *g));
extern DdNode *     extraZddPrimeProduct ARGS((DdManager *dd, DdNode *f, DdNode *g));
/* an alternative implementation of the cover product */
extern DdNode *    Extra_zddProductAlt ARGS((DdManager *dd, DdNode *f, DdNode *g));
extern DdNode *     extraZddProductAlt ARGS((DdManager *dd, DdNode *f, DdNode *g));
/* returns the set of cubes pair-wise unate with the given cube */
extern DdNode *    Extra_zddCompatible ARGS((DdManager * dd, DdNode * zCover, DdNode * zCube));
extern DdNode *     extraZddCompatible ARGS((DdManager * dd, DdNode * zCover, DdNode * zCube));
/* a wrapper for the call to Extra_zddIsop() */
extern DdNode *    Extra_zddIsopCover ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
/* a wrapper for the call to Extra_zddIsopCover() and Extra_zddPrintCover() */
extern void        Extra_zddIsopPrintCover ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
/* a simple cover computation (not ISOP) */
extern DdNode *    Extra_zddSimplify ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
extern DdNode *     extraZddSimplify ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
/* an alternative ISOP cover computation (faster than Extra_zddIsop()) */
extern DdNode *    Extra_zddIsopCoverAlt ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
extern DdNode *     extraZddIsopCoverAlt ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
/* count the number of cubes in the ISOP without building the ISOP as a ZDD */
extern int         Extra_zddIsopCubeNum ARGS((DdManager * dd, DdNode * F1, DdNode * F12));
extern DdNode *     extraZddIsopCubeNum ARGS((DdManager * dd, DdNode * F1, DdNode * F12, int * pnCubes));

/* computes the disjoint cube cover produced by the bdd paths */
extern DdNode *    Extra_zddDisjointCover ARGS((DdManager * dd, DdNode * F));
/* performs resolution on the set of clauses (S) w.r.t. variables in zdd Vars */
extern DdNode *    Extra_zddResolve ARGS((DdManager * dd, DdNode * S, DdNode * Vars));
/* cubes from zC that are not contained by cubes from zD over area bA */
extern DdNode *    Extra_zddNotContainedCubesOverArea ARGS((DdManager * dd, DdNode * zC, DdNode * zD, DdNode * bA));
extern DdNode *     extraZddNotContainedCubesOverArea ARGS((DdManager * dd, DdNode * zC, DdNode * zD, DdNode * bA));

/* finds cofactors of the cover w.r.t. the top-most variable without creating new nodes */
extern void         extraDecomposeCover ARGS((DdManager* dd, DdNode *C, DdNode **zC0, DdNode **zC1, DdNode **zC2 ));
/* composes the cover from the three subcovers using the given variable (returns NULL = reordering)*/
extern DdNode *     extraComposeCover ARGS((DdManager* dd, DdNode * zC0, DdNode * zC1, DdNode * zC2, int TopVar ));

/* selects one cube from a ZDD representing the cube cover */
extern DdNode *    Extra_zddSelectOneCube ARGS((DdManager * dd, DdNode * zS));
extern DdNode *     extraZddSelectOneCube ARGS((DdManager * dd, DdNode * zS));
/* selects one subset from a ZDD representing the set of subsets */
extern DdNode *    Extra_zddSelectOneSubset ARGS((DdManager * dd, DdNode * zS));
extern DdNode *     extraZddSelectOneSubset ARGS((DdManager * dd, DdNode * zS));

/* checks unateness of the cover */
extern int         Extra_zddCheckUnateness ARGS((DdManager * dd, DdNode * zCover));

/*=== extraZddMaxMin.c ==============================================================*/

/* maximal/minimimal */
extern DdNode *    Extra_zddMaximal ARGS((DdManager *dd, DdNode *S));
extern DdNode *     extraZddMaximal ARGS((DdManager *dd, DdNode *S));
extern DdNode *    Extra_zddMinimal ARGS((DdManager *dd, DdNode *S));
extern DdNode *     extraZddMinimal ARGS((DdManager *dd, DdNode *S));
/* maximal/minimal of the union of two sets of subsets */
extern DdNode *    Extra_zddMaxUnion ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *     extraZddMaxUnion ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *    Extra_zddMinUnion ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *     extraZddMinUnion ARGS((DdManager *dd, DdNode *S, DdNode *T));
/* dot/cross products */
extern DdNode *    Extra_zddDotProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *     extraZddDotProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *    Extra_zddExorProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *     extraZddExorProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *    Extra_zddCrossProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *     extraZddCrossProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *    Extra_zddMaxDotProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));
extern DdNode *     extraZddMaxDotProduct ARGS((DdManager *dd, DdNode *S, DdNode *T));

/*=== extraZddMisc.c ==============================================================*/

/* create the combination composed of a single ZDD variable */
extern DdNode *    Extra_zddVariable ARGS((DdManager * dd, int iVar));
/* build a ZDD for a combination of variables */
extern DdNode *    Extra_zddCombination ARGS((DdManager *dd, int* VarValues, int nVars));
extern DdNode *     extraZddCombination ARGS((DdManager *dd, int *VarValues, int nVars ));
/* the set of all possible combinations of the given set of variables */
extern DdNode *    Extra_zddUniverse ARGS((DdManager * dd, DdNode * VarSet));
extern DdNode *     extraZddUniverse ARGS((DdManager * dd, DdNode * VarSet));
/* build the set of all tuples of K variables out of N */
extern DdNode *    Extra_zddTuples ARGS((DdManager * dd, int K, DdNode *zVarsN));
extern DdNode *     extraZddTuples ARGS((DdManager * dd, DdNode *zVarsK, DdNode *zVarsN));
/* build the set of all tuples of K variables out of N from the BDD cube */
extern DdNode *    Extra_zddTuplesFromBdd ARGS((DdManager * dd, int K, DdNode *bVarsN));
extern DdNode *     extraZddTuplesFromBdd ARGS((DdManager * dd, DdNode *bVarsK, DdNode *bVarsN));
/* convert the set of singleton combinations into one combination */
extern DdNode *    Extra_zddSinglesToComb ARGS((DdManager * dd, DdNode * Singles));
extern DdNode *     extraZddSinglesToComb ARGS(( DdManager * dd, DdNode * Singles ));
/* returns the set of combinations containing the max/min number of elements */
extern DdNode *    Extra_zddMaximum ARGS((DdManager * dd, DdNode * S, int * nVars));
extern DdNode *     extraZddMaximum ARGS((DdManager * dd, DdNode * S, int * nVars));
extern DdNode *    Extra_zddMinimum ARGS((DdManager * dd, DdNode * S, int * nVars));
extern DdNode *     extraZddMinimum ARGS((DdManager * dd, DdNode * S, int * nVars));
/* returns the random set of k combinations of n elements with average density d */
extern DdNode *    Extra_zddRandomSet ARGS((DdManager * dd, int n, int k, double d));

/*=== extraZddSubSup.c ==============================================================*/

/* subset/supset operations */
extern DdNode *    Extra_zddSubSet    ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *     extraZddSubSet    ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *    Extra_zddSupSet    ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *     extraZddSupSet    ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *    Extra_zddNotSubSet ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *     extraZddNotSubSet ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *    Extra_zddNotSupSet ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *     extraZddNotSupSet ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *    Extra_zddMaxNotSupSet ARGS((DdManager *dd, DdNode *X, DdNode *Y));
extern DdNode *     extraZddMaxNotSupSet ARGS((DdManager *dd, DdNode *X, DdNode *Y));
/* check whether the empty combination belongs to the set of subsets */
extern int         Extra_zddEmptyBelongs  ARGS((DdManager *dd, DdNode* zS ));
/* check whether the set consists of one subset only */
extern int         Extra_zddIsOneSubset ARGS((DdManager *dd, DdNode* zS ));

/*===========================================================================*/
/*     Various Utilities                                                     */
/*===========================================================================*/

/*=== extraUtilPrint.c =================================================================*/

enum { 
	EXTRA_GATE_TYPE_NONE, 
	EXTRA_GATE_TYPE_BUF, 
	EXTRA_GATE_TYPE_INV, 
	EXTRA_GATE_TYPE_AND, 
	EXTRA_GATE_TYPE_NAND, 
	EXTRA_GATE_TYPE_OR, 
	EXTRA_GATE_TYPE_NOR, 
	EXTRA_GATE_TYPE_EXOR, 
	EXTRA_GATE_TYPE_NEXOR,
	EXTRA_GATE_TYPE_CROSS 
};

extern FILE *      Extra_WriteBlifStart ARGS((char * FileName, char * pInputNames[], int nInputs, char * pOutputNames[], int nOutputs));
extern void        Extra_WriteBlifStop ARGS((FILE * pFile));

/* write the BDD/ZDD into as the boolean node in the BLIF file. */
extern void        Extra_WriteBlifNode ARGS((FILE * pFile, DdManager * dd, DdNode * Func, char * pInputNames[], char * pOutputName));
extern void        Extra_WriteBlifNodeUsingGates ARGS((FILE * pFile, DdManager * dd, DdNode * Func, char * pInputNames[], char * pOutputName, int fCascade));
extern void        Extra_WriteBlifGate ARGS((FILE * pFile, char * pInputNames[], int Polars[], int nInputs, char * pOutputName, int GateType));
extern void        Extra_WriteBlifGateCascade ARGS((FILE * pFile, char * pInputNames[], int Polars[], int nInputs, char * pOutputName, int GateType));

extern unsigned    Extra_ReadBinary ARGS((char * Buffer));
extern void        Extra_PrintBinary ARGS((FILE * pFile, unsigned Sign[], int nBits));
extern void        Extra_PrintSupport ARGS((DdManager * dd, DdNode * Func));
extern void        Extra_PrintSymbols ARGS((FILE * pFile, char Char, int nTimes, int fPrintNewLine));
extern void        Extra_DumpDot ARGS((DdManager * dd, DdNode * pFuncs[], int nFuncs, char * FileName, int fFlagZdd));
extern int         Extra_FileNameCheckExtension ARGS((char * FileName, char * Extension));
extern char *      Extra_FileNameAppend ARGS((char * pBase, char * pSuffix));
extern char *      Extra_FileNameGeneric ARGS((char * FileName));
extern void        Extra_StopManager ARGS((DdManager * dd));
extern int         Extra_Power3 ARGS((int Num));
extern int *       Extra_DeriveRadixCode ARGS((int Number, int Radix, int nDigits));

/*=== extraUtilMemory.c ========================================================*/

typedef struct mm_fixed_    mm_fixed;    // fixed-size-block memory manager (belongs to Extra)
typedef struct mm_flex_     mm_flex;     // flexible-size-block memory manager (belongs to Extra)

extern mm_fixed *  memManFixedStart( int nEntrySize, int nChunkSize, int nChunksAlloc );
extern void        memManFixedStop( mm_fixed * p, int fVerbose );
extern char *      memManFixedEntryFetch( mm_fixed * p );
extern void        memManFixedEntryRecycle( mm_fixed * p, char * pEntry );
extern void        memManFixedRestart( mm_fixed * p );
extern int         memManFixedReadMemUsage( mm_fixed * p );

extern mm_flex *   memManFlexStart( int nChunkSize, int nChunksAlloc );
extern void        memManFlexStop( mm_flex * p, int fVerbose );
extern char *      memManFlexEntryFetch( mm_flex * p, int nBytes );
extern int         memManFlexReadMemUsage( mm_flex * p );

/*=== extraUtilProgress.c ================================================================*/

typedef struct ProgressBarStruct ProgressBar;

extern ProgressBar * Extra_ProgressBarStart( FILE * pFile, int nItemsTotal );
extern void        Extra_ProgressBarUpdate( ProgressBar * p, int nItemsCur );
extern void        Extra_ProgressBarStop( ProgressBar * p );

/**AutomaticEnd***************************************************************/

#endif /* __EXTRA_H__ */
